!+
! Module: emenergy_mod
!
! Authors:  J. Ballet - SAp  CEA Saclay
!          
! Description: Assign energy and quality flag to events
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module emenergy_mod

  use emutils_mod
  use cal_mos
  use eventAttributes
  use dal
  use param
  use cal
  use errorHandling

  implicit none 

  private 

! Names of the task parameters

    character(len=*), parameter :: inevent    = "ineventset", &
                                   newoutput  = "newoutput", &
                                   outevent   = "outeventset", &
                                   evflag_on  = "flagevents", &
                                   cticorr_on = "correctcti", &
                                   ccdbkg_on  = "getccdbkg", &
                                   makee_on   = "makepha", &
                                   randome1   = "randomizeenergy", &
                                   energy_on  = "correctgain", &
                                   bade3e4_on = "rejectbade3e4", &
                                   e3e4ellipse= "e3e4ellipse", &
                                   minenergy  = "minenergy", &
                                   maxperipix = "maxperipix", &
                                   fillbkg_on = "fillccdbkg", &
                                   fillthresh = "fillminnumber", &
                                   withdark   = "useccfdarkframe", &
                                   withdiag   = "withmaskedccdset", &
                                   diagfile   = "maskedccdset", &
                                   savebkg    = "writebackgroundset", &
                                   bkgfile    = "backgroundset", &
                                   tbin       = "timebin"

! Name of the extensions

    character(len=*), parameter :: event_ext  = "EVENTS", &
                                   expos_ext  = "EXPOSURE", &
                                   badpix_ext = "BADPIX", &
                                   offset_ext = "OFFSETS"

! Name of the columns

    character(len=8), parameter :: rawx = "RAWX    ", rawy = "RAWY    ", &
                                   ene1 = "ENERGYE1", ene2 = "ENERGYE2", &
                                   ene3 = "ENERGYE3", ene4 = "ENERGYE4", &
                                   patt = "PATTERN ", peri = "PERIPIX ", &
                                   time = "TIME    ", &
                                    pha = "PHA     ",   pi = "PI      "

! Routine name

    character(len=*), parameter :: taskname = "emenergy"

! Global variables
! Columns of data

    integer(kind=int8),  dimension(:), pointer   :: peripix, pattern
    integer(kind=int16), dimension(:), pointer   :: posx, posy, ener4, ener1
    integer(kind=int32), dimension(:), pointer   :: quality
    real(kind=single),   dimension(:), pointer   :: phaval, pival
    real(kind=double),   dimension(:), pointer   :: timetag
    integer(kind=int16), dimension(:,:), pointer :: offsetx, offsety

! Function flags
! onexpos   : Is the EXPOSURE extension present ?
! onevflag  : Should EVFLAG   be applied ?
! oncticorr : Should CTI_CORR be applied ?
! one4bkg   : Should CCDBKG   be applied ?
! e3e4reject: Should REJECT_E3E4 be applied ?
! onmakee   : Should MAKE_E   be applied ?
! e1random  : Should PHA be randomized ?
! onenergy  : Should ENERGY   be applied ?
! onoffvar  : Do we know the offset/variance file for EVFLAG ?
! ondark    : Do we use the CCD dark frame ?
! ondiag    : Do we use a diagnostic map ?
! onfillbkg : Do we fill the empty areas in CCDBKG ?
! outbkg    : Do we write the E4 map and timeseries for CCDBKG to a file ?
! imageMode : Is this IMAGING mode data ?
! timeMode  : Is this TIMING mode data ?
! rimagMode : Is this REDUCED IMAGING mode data ?
! ctimeMode : Is this COMPRESSED TIMING mode data ?

    logical :: onevflag, oncticorr, onmakee, onenergy
    logical :: one4bkg, ondark, ondiag, onfillbkg, e3e4reject
    logical :: onoffvar, outbkg, onexpos, e1random
    logical :: imageMode, timeMode,rimagMode, ctimeMode

! Maximum size of the read-out area
! Under/overscans (to be read in CAL)
! Window parameters

    integer :: underx, overx, undery, overy
    integer :: x_size_window, y_size_window
    integer :: windX0, windY0, windDX, windDY

  public emenergy

contains 
  
  !---------------------------------------------------------------------
  subroutine emenergy
  !---------------------------------------------------------------------

    character(len=1024) :: diagnos, backgrnd
    real :: timebin, e3cen, e4cen, e3rad, e4rad, e3ang
    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2) :: optbkg
    integer(kind=int16), dimension(:,:), pointer :: fovMap
    integer :: maxevtfill, minadubkg, perithresh

! Declarations for the DAL

    type(TableT)   :: ev_tab, ex_tab, off_tab
    type(ColumnT)  :: pha_col, pi_col

! Read parameters and initialize the CAL

    call getParamvalues(ev_tab, off_tab, diagnos, backgrnd, timebin,  &
          minadubkg, perithresh, maxevtfill, e3cen, e4cen, e3rad, e4rad, e3ang)

    if (numberOfRows(ev_tab) > 0) then

       call initCAL(ev_tab, fovMap)

       call readColumns(ev_tab, pha_col, pi_col)

       if (onevflag) then
          if (imageMode .or. rimagMode) then
                       call evflagImage(ev_tab, fovMap)
          else if (timeMode .or. ctimeMode) then
                       call evflagTime(ev_tab)
          endif
       endif

       if (ondark .or. ondiag)  &
                       call subDark(ev_tab, diagnos)

       if (one4bkg) then
                       call ccdbkg(ev_tab, off_tab, backgrnd, timebin,  &
                             minadubkg, perithresh, maxevtfill, fovMap, optbkg)
       else
                       optbkg = 0.
       endif

       if (e3e4reject) call rejectE3E4(ev_tab, optbkg,  &
                                       e3cen, e4cen, e3rad, e4rad, e3ang)

       if (onmakee)    call makeE(ev_tab, pha_col, optbkg)

       if (oncticorr)  call ctiCorr(ev_tab)

       if (onenergy)   call energy(ev_tab, pha_col, pi_col)

! Propagate keywords to the EXPOSURE extension

       if (onexpos) then
          ex_tab = table(parent(ev_tab), expos_ext)
          call setAttributes(attributable(ex_tab), attributable(ev_tab))
          call release(ex_tab)
       endif

       if (onoffvar)  call release(off_tab)

       if ((imageMode .or. rimagMode) .and. (onevflag .or. onfillbkg))  &
            call CAL_releaseMemory(fovMap)
    endif

    call release(parent(ev_tab))

  end subroutine emenergy

! Flag events next to bad pixels, borders, ... version for IMAGING mode

  subroutine evflagImage(ev_tab, fovMap)

! ev_tab    : Handle to the input/output event table
! fovMap    : Output of CAL_getFOVmap

    type(TableT), intent(in)   :: ev_tab
    integer(kind=int16), dimension(:,:), pointer :: fovMap

    integer :: irow, nrows, ibad, nbad
    integer :: xcur, ycur, ycur2, xp, yp
    integer :: windX1, windY1
   integer(kind=int32),dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2,0:1)::flag_map
    type(DataSetT) :: ev_set
    type(TableT)   :: bad_tab
    type(ColumnT)  :: badx_col, bady_col, badtype_col, badext_col, badflag_col
    integer(kind=int16), dimension(:), pointer   :: badx, bady, badtype, &
                                                    badyext, badflag
    integer(int32)                           :: curflag, testflag

    call message(AppMsg,Verbose,"Entering EVFLAG")

! flag_map(:,:,0) is the map of dead pixels and bad lines/columns (for PERIPIX)
! flag_map(:,:,1) is the map of "next to bad pixels" (for FLAG)

    flag_map = 0
    flag_map(1,:,1) = ior(flag_map(1,:,1),EVATT_CLOSE_TO_CCD_BORDER)
    flag_map(EMOS_MAX_X,:,1) = ior(flag_map(EMOS_MAX_X,:,1),  &
                                   EVATT_CLOSE_TO_CCD_BORDER)
    flag_map(:,1,1) = ior(flag_map(:,1,1),EVATT_CLOSE_TO_CCD_BORDER)
    flag_map(:,EMOS_MAX_Y,1) = ior(flag_map(:,EMOS_MAX_Y,1),  &
                                   EVATT_CLOSE_TO_CCD_BORDER)

! Reject E4 including under/overscan
! Remember that events immediately next to borders are flagged already
! Set pixels to dead if underscan=0 (true for Y)

    if (underx > 0) then
       flag_map(0,:,0) = ior(flag_map(0,:,0),EVATT_CLOSE_TO_CCD_BORDER)
    else
       flag_map(0,:,0) = ior(flag_map(0,:,0),EVATT_CLOSE_TO_DEADPIX)
    endif
    if (overx > 0) then
       flag_map(EMOS_MAX_X+1,:,0) = ior(flag_map(EMOS_MAX_X+1,:,0),  &
                                        EVATT_CLOSE_TO_CCD_BORDER)
    else
       flag_map(EMOS_MAX_X+1,:,0) = ior(flag_map(EMOS_MAX_X+1,:,0),  &
                                        EVATT_CLOSE_TO_DEADPIX)
    endif
    if (undery > 0) then
       flag_map(:,0,0) = ior(flag_map(:,0,0),EVATT_CLOSE_TO_CCD_BORDER)
    else
       flag_map(:,0,0) = ior(flag_map(:,0,0),EVATT_CLOSE_TO_DEADPIX)
    endif
    if (overy > 0) then
       flag_map(:,EMOS_MAX_Y+1,0) = ior(flag_map(:,EMOS_MAX_Y+1,0),  &
                                        EVATT_CLOSE_TO_CCD_BORDER)
    else
       flag_map(:,EMOS_MAX_Y+1,0) = ior(flag_map(:,EMOS_MAX_Y+1,0),  &
                                        EVATT_CLOSE_TO_DEADPIX)
    endif
    call setEventAttribute(ev_tab,EVATT_CLOSE_TO_CCD_BORDER)

    windX1 = windX0 + windDX - 1
    windY1 = windY0 + windDY - 1

! Special case for Binodal mode

    if (windX0-1+windDX == EMOS_MAX_X/2) then
       flag_map(EMOS_MAX_X/2,:,1) = ior(flag_map(EMOS_MAX_X/2,:,1),  &
                                        EVATT_CLOSE_TO_NODE_BOUNDARY)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_NODE_BOUNDARY)
    endif

! Special case for Window mode
! Only the left and lower side are flagged. 
! Events found on the right and upper side are perfectly OK (not truncated)

    if (windX0 > 1) then
      flag_map(windX0,:,1)=ior(flag_map(windX0,:,1),EVATT_CLOSE_TO_CCD_WINDOW)
      flag_map(windX0-1,:,0)=ior(flag_map(windX0-1,:,0),EVATT_CLOSE_TO_DEADPIX)
      call setEventAttribute(ev_tab,EVATT_CLOSE_TO_CCD_WINDOW)
    endif
    if (windY0 > 1) then
      flag_map(:,windY0,1)=ior(flag_map(:,windY0,1),EVATT_CLOSE_TO_CCD_WINDOW)
      flag_map(:,windY0-1,0)=ior(flag_map(:,windY0-1,0),EVATT_CLOSE_TO_DEADPIX)
      call setEventAttribute(ev_tab,EVATT_CLOSE_TO_CCD_WINDOW)
    endif

! Flag events outside FOV

    flag_map(windX0:windX1,windY0:windY1,1) =  &
       ior(flag_map(windX0:windX1,windY0:windY1,1),EVATT_OUT_OF_FOV*(1-fovMap))

! Offsets larger than 4000 mean the line/column is bad and deactivated

    if (onoffvar) then
       do irow = 1, EMOS_MAX_X
          if (offsetx(underx+irow-1,0) > 4000) then
           flag_map(irow,:,0)  =ior(flag_map(irow,:,0),  EVATT_CLOSE_TO_BADCOL)
           flag_map(irow-1,:,1)=ior(flag_map(irow-1,:,1),EVATT_CLOSE_TO_BADCOL)
           flag_map(irow+1,:,1)=ior(flag_map(irow+1,:,1),EVATT_CLOSE_TO_BADCOL)
          endif
       enddo
       do irow = 1, EMOS_MAX_Y
          if (offsety(undery+irow-1,0) > 4000) then
           flag_map(:,irow,0)  =ior(flag_map(:,irow,0),  EVATT_CLOSE_TO_BADROW)
           flag_map(:,irow-1,1)=ior(flag_map(:,irow-1,1),EVATT_CLOSE_TO_BADROW)
           flag_map(:,irow+1,1)=ior(flag_map(:,irow+1,1),EVATT_CLOSE_TO_BADROW)
          endif
       enddo
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_BADCOL)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_BADROW)
    endif

! Read the bad pixels extension
! Type: 0=OK, 1=hot, 2=flickering, 3=dead, 4=pinhole, 5=anomalous
! Flag: 1=uplinked, 2=in CCF but not uplinked, 3=new

    ev_set = parent(ev_tab)
    if (hasBlock(ev_set, badpix_ext)) then
       bad_tab = table(ev_set, badpix_ext)
       badx_col = column(bad_tab, "RAWX", Read)
       badx => int16Data(badx_col)
       bady_col = column(bad_tab, "RAWY", Read)
       bady => int16Data(bady_col)
       badtype_col = column(bad_tab, "TYPE", Read)
       badtype => int16Data(badtype_col)
       badext_col = column(bad_tab, "YEXTENT", Read)
       badyext => int16Data(badext_col)
       badflag_col = column(bad_tab, "BADFLAG", Read)
       badflag => int16Data(badflag_col)

       nbad = numberOfRows(bad_tab)
       do ibad = 0,nbad-1
          xcur = badx(ibad)
          ycur = bady(ibad)
          ycur2 = ycur + badyext(ibad) - 1

! Check that no bad pixel has invalid coordinates

          if (min(xcur,ycur) < 1 .or.  &
              min(EMOS_MAX_X-xcur,EMOS_MAX_Y-ycur2) < 0) then
              write(errstr,*) "invalid bad pixel coordinates", xcur, &
                              ",", ycur, ",", ycur2, " at row", ibad+1
              call warning("evflagImage13")
          else
              select case(int(badtype(ibad),kind=int32))
              case (INTACT)
                 curflag = 0
              case (DEAD)
                 curflag = EVATT_CLOSE_TO_DEADPIX
         flag_map(xcur,ycur:ycur2,0) = ior(flag_map(xcur,ycur:ycur2,0),curflag)
              case default
                 if (badflag(ibad) > 1) then
                    curflag = EVATT_CLOSE_TO_BRIGHTPIX
                 else
                    curflag = EVATT_CLOSE_TO_ONBOARD_BADPIX
                 endif
              end select

! Only pixels adjacent (not diagonal) to bad pixels are flagged

              flag_map(xcur,ycur-1,1)  = ior(flag_map(xcur,ycur-1,1),curflag)
              flag_map(xcur,ycur2+1,1) = ior(flag_map(xcur,ycur2+1,1),curflag)
     flag_map(xcur-1,ycur:ycur2,1) = ior(flag_map(xcur-1,ycur:ycur2,1),curflag)
     flag_map(xcur+1,ycur:ycur2,1) = ior(flag_map(xcur+1,ycur:ycur2,1),curflag)
          endif
       enddo
       call release(bad_tab)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_DEADPIX)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_BRIGHTPIX)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_ONBOARD_BADPIX)
    else
       write(errstr,*) "no ", badpix_ext, " extension in input event file.",  &
                       " Assume no bad pixel in EVFLAG."
       call warning("evflagImage11")
    endif

! Flag only events whose main pixel is adjacent (not diagonal) to a bad pixel.

    testflag = EVATT_CLOSE_TO_BADCOL + EVATT_CLOSE_TO_BADROW  &
             + EVATT_CLOSE_TO_CCD_BORDER  ! Flag for rejecting E4
    nrows = numberOfRows(ev_tab)
    do irow = 0, nrows-1
       if (iand(quality(irow),EVATT_OUT_OF_CCD_WINDOW) == 0) then
          xcur = posx(irow)
          ycur = posy(irow)
          curflag = ior(quality(irow),flag_map(xcur,ycur,1))
          quality(irow) = curflag

! PERIPIX is modified if a dead pixel or bad line/column is found 
! in the outer corona.
! Same if the outer corona includes the under/overscan

          curflag = 0
          do xp = -2,2
             curflag = curflag + iand(testflag,flag_map(xcur+xp,ycur,0)) &
                  + iand(testflag,flag_map(xcur,ycur+xp,0))
          enddo
          if (curflag > 0) then
             peripix(irow) = 7
          else
             curflag = 0
             do yp = -2,2
                curflag = curflag  &
                     + iand(EVATT_CLOSE_TO_DEADPIX,flag_map(xcur-2,ycur+yp,0))
                curflag = curflag  &
                     + iand(EVATT_CLOSE_TO_DEADPIX,flag_map(xcur+2,ycur+yp,0))
             enddo
             do xp = -1,1
                curflag = curflag  &
                     + iand(EVATT_CLOSE_TO_DEADPIX,flag_map(xcur+xp,ycur-2,0))
                curflag = curflag  &
                     + iand(EVATT_CLOSE_TO_DEADPIX,flag_map(xcur+xp,ycur+2,0))
             enddo
             peripix(irow) = peripix(irow) + curflag / EVATT_CLOSE_TO_DEADPIX
          endif
       endif
    enddo

    write(errstr,*) sum(iand(quality,EVATT_OUT_OF_FOV)/EVATT_OUT_OF_FOV),  &
                    " events outside field of view"
    call message(AppMsg,Sparse)
    call setEventAttribute(ev_tab,EVATT_OUT_OF_FOV)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, evflag_on)

    call message(AppMsg,Verbose,"Exiting EVFLAG")

  end subroutine evflagImage

! Flag events next to bad pixels, borders, ... 1-D version for TIMING mode

  subroutine evflagTime(ev_tab)

! ev_tab    : Handle to the input/output event table

    type(TableT), intent(in)   :: ev_tab

    integer :: irow, ibad, nbad
    integer :: xcur
    integer(kind=int32),dimension(-1:EMOS_MAX_X+2)::flag_map
    type(DataSetT) :: ev_set
    type(TableT)   :: bad_tab
    type(ColumnT)  :: badx_col, badtype_col, badflag_col
    integer(kind=int16), dimension(:), pointer   :: badx, badtype, badflag
    integer(int32)                           :: curflag

    call message(AppMsg,Verbose,"Entering EVFLAG")

! flag_map is the map of "next to bad pixels" (for FLAG)

    flag_map = 0

    flag_map(1) = ior(flag_map(1),EVATT_CLOSE_TO_CCD_BORDER)
    flag_map(EMOS_MAX_X) = ior(flag_map(EMOS_MAX_X),EVATT_CLOSE_TO_CCD_BORDER)

! Special case for Window mode
! Only the left and lower side are flagged. 
! Events found on the right and upper side are perfectly OK (not truncated)

    if (windX0 > 1)  then
      flag_map(windX0) = ior(flag_map(windX0),EVATT_CLOSE_TO_CCD_WINDOW)
      call setEventAttribute(ev_tab,EVATT_CLOSE_TO_CCD_WINDOW)
    endif

! Offsets larger than 4000 mean the column is bad and deactivated

    if (onoffvar) then
       do irow = 1, EMOS_MAX_X
          if (offsetx(underx+irow-1,0) > 4000) then
             flag_map(irow-1) = ior(flag_map(irow-1),EVATT_CLOSE_TO_BADCOL)
             flag_map(irow+1) = ior(flag_map(irow+1),EVATT_CLOSE_TO_BADCOL)
          endif
       enddo
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_BADCOL)
    endif

! Read the bad pixels extension
! Type: 0=OK, 1=hot, 2=flickering, 3=dead, 4=pinhole, 5=anomalous
! Flag: 1=uplinked, 2=in CCF but not uplinked, 3=new

    ev_set = parent(ev_tab)
    if (hasBlock(ev_set, badpix_ext)) then
       bad_tab = table(ev_set, badpix_ext)
       badx_col = column(bad_tab, "RAWX", Read)
       badx => int16Data(badx_col)
       badtype_col = column(bad_tab, "TYPE", Read)
       badtype => int16Data(badtype_col)
       badflag_col = column(bad_tab, "BADFLAG", Read)
       badflag => int16Data(badflag_col)

       nbad = numberOfRows(bad_tab)
       do ibad = 0,nbad-1
          xcur = badx(ibad)

! Check that no bad pixel has invalid coordinates

          if (xcur < 1 .or.  xcur > EMOS_MAX_X) then
             write(errstr,*) "invalid bad pixel coordinates", xcur, &
                             ", at row", ibad+1
             call warning("evflagTime13")
          else
             select case(int(badtype(ibad),kind=int32))
             case (INTACT)
                curflag = 0
             case (DEAD)
                curflag = EVATT_CLOSE_TO_DEADPIX
             case default
                if (badflag(ibad) > 1) then
                   curflag = EVATT_CLOSE_TO_BRIGHTPIX
                else
                   curflag = EVATT_CLOSE_TO_ONBOARD_BADPIX
                endif
             end select

! Only pixels adjacent (not diagonal) to bad pixels are flagged

             flag_map(xcur-1) = ior(flag_map(xcur-1),curflag)
             flag_map(xcur+1) = ior(flag_map(xcur+1),curflag)
          endif
       enddo
       call release(bad_tab)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_DEADPIX)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_BRIGHTPIX)
       call setEventAttribute(ev_tab,EVATT_CLOSE_TO_ONBOARD_BADPIX)
    else
       write(errstr,*) "no ", badpix_ext, " extension in input event file.",  &
                       " Assume no bad pixel in EVFLAG."
       call warning("evflagTime11")
    endif

! Apply flags to events

    where (iand(quality,EVATT_OUT_OF_CCD_WINDOW) == 0)
       quality = ior(quality,flag_map(posx))
    endwhere

! Write keywords to sum up action

    call setEventAttribute(ev_tab,EVATT_CLOSE_TO_CCD_BORDER)
    call keywordDone(block(ev_tab), taskname, evflag_on)

    call message(AppMsg,Verbose,"Exiting EVFLAG")

  end subroutine evflagTime

! Correct the energies for small scale variations on the CCD
! Should propagate decimals to E1 if dark frame gets real

  subroutine subDark(ev_tab, diagnos)

! ev_tab     : Handle to the input/output event table
! diagnos    : Name of the diagnostic map file

    character(len=*), intent(in) :: diagnos
    type(TableT),     intent(in) :: ev_tab

    real, parameter :: epsilon = 1.e-5
    real            :: loc2, loc3, loc4
    integer         :: irow, ix, iy, kx, ky
    type(ColumnT)   :: e2_col, e3_col
    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2) :: optbkg
    integer(kind=int8),  dimension(0:31)             :: pixelInE2
    integer(kind=int8), dimension(0:31,-1:1,-1:1)    :: patabove
    integer(kind=int16), dimension(:), pointer       :: ener2, ener3

! subDark may be called only once

    if (wasDone(block(ev_tab),"subdark")) then
       call warning("subDark10",  &
                    "SUBDARK may be called only once. Not repeated.")
       return
    endif

    call message(AppMsg,Verbose,"Entering SUBDARK")

    optbkg = 0.

    if (ondark) then
       call readDark(ev_tab, optbkg)
    else if (ondiag) then
       call readBkg(diagnos, ev_tab, optbkg)
    endif

    if (.not. (ondark .or. ondiag))  return

! Subtract offsets if diagnostic or CAL map, except if value 0

    if (onoffvar) then
       do irow = 1, EMOS_MAX_X
          where (optbkg(irow,:) > epsilon)
             optbkg(irow,:) = optbkg(irow,:) - offsetx(underx+irow-1,0)
          endwhere
       enddo
       do irow = 1, EMOS_MAX_Y
          where (optbkg(:,irow) > epsilon)
             optbkg(:,irow) = optbkg(:,irow) - offsety(undery+irow-1,0)
          endwhere
       enddo
    endif

! Read columns not declared as global variables

    e2_col = column(ev_tab, ene2, Read)
    ener2 => int16Data(e2_col)
    if (imageMode) then
       e3_col = column(ev_tab, ene3, Read)
       ener3 => int16Data(e3_col)
    endif

! Get pattern geometry

    call getpixelInE2(pixelInE2, patabove)

! Loop over all events

    do irow = 0, numberOfRows(ev_tab)-1
       if (iand(quality(irow),EVATT_OUT_OF_CCD_WINDOW) == 0) then
          ix = posx(irow)
          iy = posy(irow)

! Correct E1

          ener1(irow) = ener1(irow) - optbkg(ix,iy)

! Correct E2 and E3

          loc2 = 0.
          loc3 = 0.
          do ky = -1, 1
             do kx = -1, 1
                if (patabove(pattern(irow),kx,ky) > 0) then
                   loc2 = loc2 + optbkg(ix+kx,iy+ky)
                else
                   loc3 = loc3 + optbkg(ix+kx,iy+ky)
                endif
             enddo
          enddo
          ener2(irow) = ener2(irow) - int(loc2+0.5,kind=int16)

          if (imageMode) then
             loc4 = sum(optbkg(ix-2:ix+2,iy-2:iy+2)) - loc2 - loc3
             loc3 = loc3 - optbkg(ix,iy)   ! Central pixel is at 0 in patabove
             ener3(irow) = ener3(irow) - int(loc3+0.5,kind=int16)

! Correct E4

             loc4 = loc4 * (1.-peripix(irow)/16.)
             ener4(irow) = ener4(irow) - int(loc4+0.5,kind=int16)
          endif
       endif
    enddo

! Release columns not declared as global variables

    call release(e2_col)
    if (imageMode)  call release(e3_col)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, "subdark")
    call message(AppMsg,Verbose,"Exiting SUBDARK")

  end subroutine subDark

! Compute the optical background (energy offset) on the CCD
! The current version works in RAWX/RAWY coordinates

  subroutine ccdbkg(ev_tab, off_tab, backgrnd, timebin, minadubkg,  &
                    perithresh, maxevtfill, fovMap, optbkg)

! ev_tab     : Handle to the input/output event table
! off_tab    : Handle to the offsets table
! backgrnd   : Name of the E4 map and timeseries file
! timebin    : Time bin for E4 timeseries
! minadubkg  : Min E1 for building E4 map
! perithresh : Max PERIPIX for building E4 map
! maxevtfill : Max number of evts for averaging E4 map
! fovMap     : Output of CAL_getFOVmap
! optbkg     : Optical background on the CCD

    character(len=*), intent(in) :: backgrnd
    real,             intent(in) :: timebin
    integer,          intent(in) :: maxevtfill, minadubkg, perithresh
    integer(kind=int16), dimension(:,:), pointer :: fovMap
    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(out) :: optbkg
    type(TableT),     intent(in) :: ev_tab, off_tab

    real, parameter    :: epsilon = 1.e-5
    integer :: irow, x, y
    integer, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2) :: sigmaper
    integer(kind=int32) :: testflag
    integer(kind=int8), dimension(0:31)           :: pixelInE2

    call message(AppMsg,Verbose,"Entering CCDBKG")

!  Construct bit mask for testing the flags
!  Only 0, DIAGONAL and OUT_OF_FOV are accepted
!  Large events are not

    testflag = NOT(EVATT_DIAGONAL+EVATT_OUT_OF_FOV)

    call getpixelInE2(pixelInE2)

! Accumulate E4 and 16-PERIPIX
! Keep only PERIPIX <= maxperipix and E1 >= minenergy

    call setAttribute(ev_tab, "BKGE1MIN", minadubkg,  &
                  comment="Minimum ENERGYE1 for using event in background map")
    call setAttribute(ev_tab, "BKGPERIP", perithresh,  &
                  comment="Maximum PERIPIX for using event in background map")

    optbkg   = 0.
    sigmaper = 0
    do irow = 0, numberOfRows(ev_tab)-1

! Test must be the same for time-series in writeBkg

       if (ener1(irow) < minadubkg .or. iand(quality(irow),testflag) > 0  &
           .or. pixelInE2(pattern(irow)) > 1  &
           .or. peripix(irow) > perithresh)  cycle
       x = posx(irow)
       y = posy(irow)
       optbkg(x-2:x+2,y-2:y+2)   = optbkg(x-2:x+2,y-2:y+2) + ener4(irow)
       sigmaper(x-2:x+2,y-2:y+2) = sigmaper(x-2:x+2,y-2:y+2)  &
                                 + (16-peripix(irow))
    enddo
    where (sigmaper /= 0)
       optbkg = optbkg / sigmaper
    endwhere

! Fill optbkg with row and column average, separately in and out of FOV

    if (onfillbkg) then
       call fillCcdBkg(optbkg, sigmaper, fovMap, maxevtfill)
       fovMap = 1 - fovMap
       call fillCcdBkg(optbkg, sigmaper, fovMap, maxevtfill)
       fovMap = 1 - fovMap
       call keywordDone(block(ev_tab), taskname, fillbkg_on)
       call setAttribute(ev_tab, "NMAXFILL", maxevtfill,  &
                         comment="Background map is filled when less events")
    endif

    if (outbkg)  call writeBkg(backgrnd, ev_tab, off_tab, optbkg, sigmaper,  &
                  timebin, testflag, minadubkg, perithresh, pixelInE2, epsilon)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, ccdbkg_on)
    call message(AppMsg,Verbose,"Exiting CCDBKG")

  end subroutine ccdbkg

! Analyse charge below threshold (E3, E4)

  subroutine rejectE3E4(ev_tab, optbkg, e3cen, e4cen, e3rad, e4rad, e3ang)

! ev_tab    : Handle to the event table
! optbkg    : Optical background on the CCD
! e3cen     : Center of ellipse along E3/E1
! e4cen     : Center of ellipse along E4/E1
! e3rad     : Radius of ellipse along first axis
! e4rad     : Radius of ellipse along second axis
! e3ang     : Angle of first axis with E3 axis

    type(TableT), intent(in) :: ev_tab
    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(in) :: optbkg
    real,         intent(in) :: e3cen, e4cen, e3rad, e4rad, e3ang

    type(ColumnT)  :: e3_col
    integer(kind=int8), dimension(0:31)        :: pixelInE2
    integer(kind=int16), dimension(:), pointer :: ener3
    integer, dimension(:), pointer :: offX, offY
    real    :: locbkg, loce1, loce3, loce4, ce3ang, se3ang
    integer :: numb, irow, xori, yori
    logical :: newcoords

    call message(AppMsg,Verbose,"Entering REJECT_E3E4 ...")

    e3_col = column(ev_tab, ene3, Read)
    ener3 => int16Data(e3_col)

! Do not reject bad E3 or E4 due to lines/columns with offsets set to 4095
! 1 means at most 2 pixels away from bad line/column

    allocate(offX(-2:x_size_window+1),offY(-2:y_size_window+1))
    offX = 0
    offY = 0
    if (onoffvar) then
       do irow = 0, x_size_window-1
          if (offsetx(irow,0) > 4000)  &
             offX(irow-2:irow+2) = max(offX(irow-2:irow+2),1)
       enddo
       do irow = 0, y_size_window-1
          if (offsety(irow,0) > 4000)  &
             offY(irow-2:irow+2) = max(offY(irow-2:irow+2),1)
       enddo
    endif

    call getpixelInE2(pixelInE2)
    ce3ang = cos(e3ang)
    se3ang = sin(e3ang)

    newcoords = wasDone(block(ev_tab),changexy)
    numb = sum(iand(quality,EVATT_BAD_E3E4)/EVATT_BAD_E3E4)

    do irow = 0, size(ener1)-1
       if (iand(quality(irow),EVATT_BAD_E3E4) > 0)  cycle
       if (newcoords) then
          if (iand(quality(irow),EVATT_OUT_OF_CCD_WINDOW) > 0)  cycle
          xori = posx(irow) + underx - 1
          yori = posy(irow) + undery - 1
       else
          xori = max(min(posx(irow)-2,x_size_window-1),0)
          yori = max(min(posy(irow)-2,y_size_window-1),0)
       endif
       if (offX(xori) + offY(yori) >= 1)  cycle
       locbkg = optbkg(xori-underx+1,yori-undery+1)
       loce1 = ener1(irow) - locbkg
       loce3 = ener3(irow) - locbkg * (8-pixelInE2(pattern(irow)))
       loce4 = ener4(irow) - locbkg * (16-peripix(irow))
       loce3 = loce3/loce1 - e3cen
       loce4 = loce4/loce1 - e4cen
       if (((loce3*ce3ang+loce4*se3ang)/e3rad)**2 +  &
           ((loce4*ce3ang-loce3*se3ang)/e4rad)**2 > 1.)  &
           quality(irow) = ior(quality(irow),EVATT_BAD_E3E4)
    enddo

    numb = sum(iand(quality,EVATT_BAD_E3E4)/EVATT_BAD_E3E4) - numb
    write(errstr,*) numb, ' events outside the ellipse in E3/E1, E4/E1'
    call message(AppMsg,Sparse)

    deallocate(offX, offY)
    call release(e3_col)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, bade3e4_on)
    call setEventAttribute(ev_tab,EVATT_BAD_E3E4)
    call setAttribute(ev_tab, "E3CEN", e3cen,  &
                      comment="Centering of E3/E1 in rejection ellipse")
    call setAttribute(ev_tab, "E4CEN", e4cen,  &
                      comment="Centering of E4/E1 in rejection ellipse")
    call setAttribute(ev_tab, "E3RAD", e3rad,  &
                      comment="Radius along first axis of rejection ellipse")
    call setAttribute(ev_tab, "E4RAD", e4rad,  &
                      comment="Radius along second axis of rejection ellipse")
    call setAttribute(ev_tab, "E3ANG", e3ang, "rad",  &
                      comment="Angle of first ellipse axis with E3/E1 axis")

  end subroutine rejectE3E4

! Compute PHA from E1, E2, E3, E4

  subroutine makeE(ev_tab, pha_col, optbkg)

! ev_tab    : Handle to the input/output event table
! pha_col   : Handle to the PHA column
! optbkg    : Optical background on the CCD

    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(in) :: optbkg
    type(TableT), intent(in)   :: ev_tab
    type(ColumnT), intent(in)  :: pha_col

    type(ColumnT) :: e2_col, e3_col
    integer(kind=int16), dimension(:), pointer   :: ener2, ener3, edum
    integer :: nrows, irow, ix, iy, nabove, phamax, phamin
    real(kind=single),   dimension(:), pointer   :: e1rand, locbkg

    call message(AppMsg,Verbose,"Entering MAKE_E")
    nrows = numberOfRows(ev_tab)

    if (imageMode .or. rimagMode) then

! Randomize ener1

       call CAL_integerToReal(ener1,e1rand)

       e2_col = column(ev_tab, ene2, Read)
       ener2 => int16Data(e2_col)
       if (imageMode) then
          e3_col = column(ev_tab, ene3, Read)
          ener3 => int16Data(e3_col)
       endif

       if (one4bkg) then

! Allocate the local background array
! Get local background for each event
! Do not use non-standard Integer as index

          allocate(locbkg(0:size(phaval)-1))
          do irow = 0, nrows-1
             if (iand(quality(irow),EVATT_OUT_OF_CCD_WINDOW) == 0) then
                ix = posx(irow)
                iy = posy(irow)
                locbkg(irow) = optbkg(ix,iy)
             else
                locbkg(irow) = 0
             endif
          enddo

          if (imageMode) then
             call CAL_mosPhaBuild(e1rand, ener2, pattern, phaval,  &
                                  locbkg, ener3, ener4, peripix, quality)
          else
             call CAL_mosPhaBuild(e1rand, ener2, pattern, phaval,  &
                                  locbkg=locbkg)
          endif

       else

! No local background

          if (imageMode) then
             call CAL_mosPhaBuild(e1rand, ener2, pattern, phaval,  &
                energye3=ener3, energye4=ener4, peripix=peripix, flag=quality)
          else
             call CAL_mosPhaBuild(e1rand, ener2, pattern, phaval)
          endif
       endif

       call release(e2_col)
       call CAL_releaseMemory(e1rand)
       if (imageMode)  call release(e3_col)
       if (one4bkg)  deallocate(locbkg)

! Range check (0.5 tolerance because pha is real)

       if (hasAttribute(pha_col,"TDMIN")) then
          phamin = int32Attribute(pha_col,"TDMIN")
       else
          phamin = 0
       endif
       if (hasAttribute(pha_col,"TDMAX")) then
          phamax = int32Attribute(pha_col,"TDMAX")
       else
          phamax = 32767
       endif
       nabove = 0
       do irow = 0, nrows-1
          if (phaval(irow) < phamin-0.5 .or. phaval(irow) > phamax+0.5) then
             nabove = nabove + 1
             quality(irow) = ior(quality(irow),EVATT_OUTSIDE_THRESHOLDS)
          endif
       enddo
       write(errstr,*) nabove, " events with PHA out of range",  &
                       phamin, " -", phamax
       call message(AppMsg,Sparse)
       call setAttribute(pha_col, "TLMIN",      0,  &
                         comment="Legal minimum for "//pha)
       call setAttribute(pha_col, "TLMAX",  32767,  &
                         comment="Legal maximum for "//pha)

    else if (timeMode .or. ctimeMode) then

! Apply randomisation 

       if (e1random) then
          allocate(edum(0:size(phaval)-1))
          edum = int(phaval+0.5,kind=int16)
          call CAL_integerToReal(edum,phaval)
          deallocate(edum)
       endif
    endif

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, makee_on)
    if (e1random) then
       call setAttribute(ev_tab, "RAND_PI", 1,  &
                         comment="Energy values are randomized over 1 ADU")
    else
       call setAttribute(ev_tab, "RAND_PI", 0,  &
                         comment="Energy values are not randomized")
    endif

    call message(AppMsg,Verbose,"Exiting MAKE_E")

  end subroutine makeE

! Correct for charge transfer inefficiency (CAL call). Write result in PI.

  subroutine ctiCorr(ev_tab)

! ev_tab    : Handle to the input/output event table

    type(TableT), intent(in)   :: ev_tab

    integer(kind=int16), dimension(:), pointer   :: dumx, dumy

    call message(AppMsg,Verbose,"Entering CTI_CORR")

! CTI_CORR resets the PI column

    if (wasDone(block(ev_tab),energy_on))  &
       call keywordRemove(block(ev_tab),energy_on)

! Remove the offending values of RAWX/RAWY

    allocate(dumx(0:size(posx)-1),dumy(0:size(posy)-1))
    where (iand(quality,EVATT_OUT_OF_CCD_WINDOW) == 0)
       dumx = posx
       dumy = posy
    elsewhere
       dumx = 1
       dumy = 1
    endwhere
    call CAL_mosCtiCorrect(timetag, dumx, dumy, phaval, pival)
    deallocate(dumx,dumy)

! Write keywords to sum up action

    call keywordDone(block(ev_tab), taskname, cticorr_on)

    call message(AppMsg,Verbose,"Exiting CTI_CORR")

  end subroutine ctiCorr

! Compute PI from CTI corrected PHA (CAL call)

  subroutine energy(ev_tab, pha_col, pi_col)

! ev_tab    : Handle to the input/output event table
! pha_col   : Handle to the PHA column
!  pi_col   : Handle to the  PI column

    type(TableT), intent(in)   :: ev_tab
    type(ColumnT), intent(in)  :: pha_col, pi_col

    real(single), dimension(:), pointer   :: dummy
    real(single),  dimension(0:1) :: piext
    real(double),  dimension(0:1) :: timeext
    integer(int8), dimension(0:1) :: pattext
    integer :: irow, pimax, pimin, nabove

    call message(AppMsg,Verbose,"Entering ENERGY")

! Reset PI to PHA if CTI_CORR is off
    if (.not. oncticorr) then
       pival = phaval
       if (wasDone(block(ev_tab),cticorr_on))  &
          call keywordRemove(block(ev_tab),cticorr_on)
    endif

    call CAL_mosGainCorrect(timetag, pattern, pival, dummy)
    pival = dummy
    call CAL_releaseMemory(dummy)

! Check in range (0.5 tolerance because pi is real)

    if (hasAttribute(pha_col,"TDMIN")) then
       pimin = int32Attribute(pha_col,"TDMIN")
    else
       pimin = 0
    endif
    if (hasAttribute(pha_col,"TDMAX")) then
       pimax = int32Attribute(pha_col,"TDMAX")
    else
       pimax = 4095
    endif

    piext(0) = pimin
    piext(1) = pimax
    timeext(0) = timetag(0)
    timeext(1) = timetag(0)
    pattext = 0
    call CAL_mosGainCorrect(timeext, pattext, piext, dummy)
    pimin = nint(dummy(0))
    pimax = nint(dummy(1))
    call CAL_releaseMemory(dummy)

    nabove = 0
    do irow = 0, numberOfRows(ev_tab)-1
       if (pival(irow) < pimin-0.5 .or. pival(irow) > pimax+0.5) then
          nabove = nabove + 1
          quality(irow) = ior(quality(irow),EVATT_OUTSIDE_THRESHOLDS)
       endif
    enddo
    write(errstr,*) nabove," events with PI out of range ",pimin," - ",pimax
    call message(AppMsg,Verbose)

! Write keywords to sum up action

    call setAttribute(pi_col, "TLMIN",     0, comment="Legal minimum for "//pi)
    call setAttribute(pi_col, "TLMAX", 15000, comment="Legal maximum for "//pi)
    call setAttribute(pi_col, "TDMIN", pimin,comment="Actual minimum for "//pi)
    call setAttribute(pi_col, "TDMAX", pimax,comment="Actual maximum for "//pi)
    call keywordDone(block(ev_tab), taskname, energy_on)

    call message(AppMsg,Verbose,"Exiting ENERGY")

  end subroutine energy

! Fill  areas <  maxevtfill evts in background map using row and column average
! Leave areas >= maxevtfill evts in background map untouched (sources)
! Only regions where fovMap = 1 are considered

  subroutine fillCcdBkg(optbkg, sigmaper, fovMap, maxevtfill)

! optbkg     : Optical background on the CCD
! sigmaper   : map of weights
! fovMap     : Output of CAL_getFOVmap
! maxevtfill : Max number of evts for averaging E4 map

    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(inout) :: optbkg
    integer, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(in) :: sigmaper
    integer(kind=int16), dimension(:,:), pointer :: fovMap
    integer, intent(in) :: maxevtfill

    integer, dimension(:,:), allocatable :: mask
    real,    dimension(:),   allocatable :: prox
    integer, dimension(:),   allocatable :: wx
    real,    dimension(:),   allocatable :: proy
    integer, dimension(:),   allocatable :: wy
    real    :: aver
    integer :: irow, windX1, windY1

    allocate(mask(windDX,windDY), prox(windDX), wx(windDX),  &
                                  proy(windDY), wy(windDY))
    windX1 = windX0 + windDX - 1
    windY1 = windY0 + windDY - 1

! Identify areas where less than maxevtfill events contribute

    mask = sigmaper(windX0:windX1,windY0:windY1) / (16*maxevtfill)
    mask = (1 - min(mask,1)) * fovMap

! Build average background per column
! Keep only columns where more than maxevtfill events contribute

    prox = sum(optbkg(windX0:windX1,windY0:windY1)  &
             * sigmaper(windX0:windX1,windY0:windY1) * mask,2)
      wx = sum(sigmaper(windX0:windX1,windY0:windY1) * mask,2)
    aver = sum(wx)
    if ( aver > 0.5 )  aver = sum(prox) / aver
    where ( wx < 16*maxevtfill )
       prox = 0
    elsewhere
       prox = prox / wx - aver
    endwhere

! Build average background per row
! Keep only rows where more than maxevtfill events contribute

    proy = sum(optbkg(windX0:windX1,windY0:windY1)  &
             * sigmaper(windX0:windX1,windY0:windY1) * mask,1)
      wy = sum(sigmaper(windX0:windX1,windY0:windY1) * mask,1)
    where ( wy < 16*maxevtfill )
       proy = 0
    elsewhere
       proy = proy / wy - aver
    endwhere

! Fill empty areas with prox + proy

    where (mask > 0)
       optbkg(windX0:windX1,windY0:windY1) = aver
    endwhere
    do irow = 1, windDX
       optbkg(irow+windX0-1,windY0:windY1) =  &
            optbkg(irow+windX0-1,windY0:windY1) + prox(irow)*mask(irow,:)
    enddo
    do irow = 1, windDY
       optbkg(windX0:windX1,irow+windY0-1) =  &
            optbkg(windX0:windX1,irow+windY0-1) + proy(irow)*mask(:,irow)
    enddo

    deallocate(mask, prox, wx, proy, wy)

  end subroutine fillCcdBkg

! Read CCD dark frame from the CCF

  subroutine readDark(ev_tab, optbkg)

! ev_tab    : Handle to the input/output event table
! optbkg    : Dark frame for the CCD

    type(TableT), intent(in)     :: ev_tab
    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(out) :: optbkg

    real(kind=single),   dimension(:,:), pointer :: darkCcf
    integer(kind=int16), dimension(:),   pointer :: caloffX, caloffY
    integer           :: windX0, windDX, windX1, windY0, windDY, windY1, irow
    character(len=80) :: shortname

    shortname = "Reference for " // taskname // "/subdark CCF" 
    call addHistory(ev_tab, shortname)

    call CAL_getMOSdarkFrameMap(darkCcf)
    optbkg(1:EMOS_MAX_X,1:EMOS_MAX_Y) = darkCcf
    call CAL_releaseMemory(darkCcf)

! Subtract offsets from CAL if no OFFSETS extension in events file

    if (.not. onoffvar) then
       write(errstr,*) "Taking offsets for the CCD dark frame from the CAL."
       call message(AppMsg,Verbose)

! Read the window size

       windX0 = int32Attribute(ev_tab,"WINDOWX0")
       windDX = int32Attribute(ev_tab,"WINDOWDX")
       windX1 = windX0 + windDX - 1
       windY0 = int32Attribute(ev_tab,"WINDOWY0")
       windDY = int32Attribute(ev_tab,"WINDOWDY")
       windY1 = windY0 + windDY - 1

! Read the offsets from the CAL

       call CAL_getMOSoffsets(caloffY,caloffX)
       if (size(caloffX) /= windDX .or. size(caloffY) /= windDY) then
          write(errstr,*) "The window size in ", trim(qualifiedName(ev_tab)), &
             " ", windDX, windDY, "is not the same as the size ",  &
             "of the arrays returned by CAL_getMOSoffsets ",  &
             size(caloffX), size(caloffY), ". Contact developer"
          call warning("readDark11")
          call CAL_releaseMemory(caloffX,caloffY)
          ondark = .false.
          optbkg = 0.
          return
       endif

       do irow = windX0, windX1
          optbkg(irow,:) = optbkg(irow,:) - caloffX(irow-windX0)
       enddo
       do irow = windY0, windY1
          optbkg(:,irow) = optbkg(:,irow) - caloffY(irow-windY0)
       enddo

       call CAL_releaseMemory(caloffX,caloffY)
    endif

  end subroutine readDark

! Read CCD dark frame from a file

  subroutine readBkg(diagnos, ev_tab, optbkg)

! diagnos   : Name of the diagnostic map file
! ev_tab    : Handle to the input/output event table
! optbkg    : Dark frame for the CCD

    character(len=*), intent(in) :: diagnos
    type(TableT), intent(in)     :: ev_tab
    real, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(out) :: optbkg

! Keywords to check for compatibility of diagnostic map and event files

    character(len=8), dimension(4), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","CCDID   ","CCDNODE "/)

    type(DataSetT) :: e4_set
    type(ArrayT)   :: opt_arr
    type(TableT)   :: off_tab
    type(ColumnT)  :: offx_col, offy_col
    integer,             dimension(:), pointer   :: sizarray
    integer(kind=int16), dimension(:,:), pointer :: ccdimg
    real(kind=single),   dimension(:,:), pointer :: e4img
    integer :: enddir, irow
    character(len=80) :: shortname

    write(errstr,*) "Reading reference CCD dark frame from ", trim(diagnos)
    call message(AppMsg,Verbose)
    e4_set = dataSet(diagnos, Read)
    opt_arr = array(e4_set, "PRIMARY", Read)
    if (equalKeywords(attributable(opt_arr),  &
                      attributable(ev_tab),keycompat)) then
       if (numberOfDimensions(opt_arr) == 2) then
          sizarray => dimensions(opt_arr)
          if (min(sizarray(0)-EMOS_MAX_X,sizarray(1)-EMOS_MAX_Y) >= 0) then
             select case(dataType(opt_arr))
             case (Integer16)
                ccdimg => int16Array2Data(opt_arr)
                optbkg(1:EMOS_MAX_X,1:EMOS_MAX_Y) = ccdimg
             case (Real32)
                e4img => real32Array2Data(opt_arr)
                optbkg(1:EMOS_MAX_X,1:EMOS_MAX_Y) = e4img
             case default
                write(errstr,*) "Unexpected array data type =",  &
                   dataType(opt_arr), " in ", trim(name(e4_set)),  &
                   ". Should be Integer16 or Real32. Not used."
                call warning("readBkg14")
                ondiag = .false.
             endselect
            if (max(sizarray(0)-EMOS_MAX_X,sizarray(1)-EMOS_MAX_Y) > 0) then
                write(errstr,*) "The array in ", trim(name(e4_set)),  &
                                " has dimensions", sizarray, " larger than",  &
                                EMOS_MAX_X, EMOS_MAX_Y, ". Used anyway."
                call warning("readBkg12")
             endif
          else
             write(errstr,*) "The array in ", trim(name(e4_set)),  &
                             " has dimensions", sizarray, " less than",  &
                             EMOS_MAX_X, EMOS_MAX_Y, ". Not used."
             call warning("readBkg11")
             ondiag = .false.
          endif
       else
          write(errstr,*) "The array in ", trim(name(e4_set)),  &
                          " is not 2-D. Not used."
          call warning("readBkg10")
          ondiag = .false.
       endif
    else
       write(errstr,*) "The CCD dark frame in ", trim(name(e4_set)),  &
           " is not compatible with the event file. File not used."
       call warning("readBkg16")
    endif
    call release(opt_arr)

! Subtract offsets from file if no OFFSETS extension in events file

    if (.not.onoffvar .and. hasBlock(e4_set,offset_ext)) then
       write(errstr,*) "Taking offsets for the CCD dark frame ",  &
                       "from the file itself"
       call message(AppMsg,Verbose)
       off_tab = table(e4_set, offset_ext)
       offx_col = column(off_tab, "COLOFFST", Read)
       offsetx => int16VectorData(offx_col)
       offy_col = column(off_tab, "ROWOFFST", Read)
       offsety => int16VectorData(offy_col)
       do irow = 1, EMOS_MAX_X
          optbkg(irow,:) = optbkg(irow,:) - offsetx(underx+irow-1,0)
       enddo
       do irow = 1, EMOS_MAX_Y
          optbkg(:,irow) = optbkg(:,irow) - offsety(undery+irow-1,0)
       enddo
    else if (.not.onoffvar) then
       write(errstr,*) "No offsets were found in the events file ",  &
          "nor in the masked CCD file. Dark Frame not applied."
       call warning("readBkg13")
       ondiag = .false.
       optbkg = 0.
    endif
    call release(e4_set)
    if (ondiag) then
       enddir = index(diagnos,'/',.true.)
       shortname = "Reference for " // taskname // "/ccdbkg " &
                 // diagnos(enddir+1:len_trim(diagnos))
       call addHistory(ev_tab, shortname)
    endif

  end subroutine readBkg

! Write the map and timeseries to a file

  subroutine writeBkg(backgrnd, ev_tab, off_tab, optbkg, sigmaper, timebin,  &
                      testflag, minadubkg, perithresh, pixelInE2, epsilon)

! backgrnd   : Name of the E4 map and timeseries file
! ev_tab     : Handle to the input/output event table
! off_tab    : Handle to the offsets table
! optbkg     : Optical background on the CCD
! sigmaper   : Weight of each value in optbkg
! timebin    : Time bin for E4 timeseries
! testflag   : Bit map for testing the flags
! minadubkg  : Min E1 for building E4 map
! perithresh : Max PERIPIX for building E4 map
! pixelInE2  : Number of pixels above threshold in inner corona
! epsilon    : small real number

    character(len=*),    intent(in) :: backgrnd
    type(TableT),        intent(in) :: ev_tab, off_tab
    real,    dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(in) :: optbkg
    integer, dimension(-1:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2), intent(in) :: sigmaper
    real,                intent(in) :: timebin, epsilon
    integer(kind=int32), intent(in) :: testflag
    integer(kind=int8), dimension(0:31), intent(in) :: pixelInE2
    integer,             intent(in) :: minadubkg, perithresh

    integer, parameter :: nbinmax = 1000000      ! Maximum size of time-series

    type(DataSetT) :: e4_set
    type(TableT)   :: e4_tab, off2_tab
    type(ArrayT)   :: opt_arr, wght_arr
    type(ColumnT)  :: te4_col, tper_col, offx_col, offy_col
    integer(kind=int32), dimension(:), pointer   :: timeper
    real(kind=single),   dimension(:), pointer   :: timee4
    real(kind=single),   dimension(:,:), pointer :: e4img
    integer(kind=int32), dimension(:,:), pointer :: e4wght
    character(len=80) :: shortname, comm
    integer :: enddir, irow, nbin, ibin
    real(kind=double) :: time0

    write(errstr,*) "Write background file ", trim(backgrnd)
    call message(AppMsg,Verbose)
    e4_set = dataSet(backgrnd, Create)
    call setAttributes(attributable(e4_set),attributable(parent(ev_tab)))
    shortname = "Space and time distribution of the E4 data"
    call addComment(e4_set, shortname)
    shortname = name(parent(ev_tab))
    enddir = index(shortname,'/',.true.)
    comm = "Input event file " &
           // shortname(enddir+1:len_trim(shortname))
    call addHistory(e4_set, comm)
    opt_arr = addArray(e4_set, "PRIMARY", Real32, (/EMOS_MAX_X,EMOS_MAX_Y/),  &
                       "space variations of background E4")
    call setAttributes(attributable(opt_arr), attributable(ev_tab))
    call deleteAttribute(opt_arr,"EXTVER")
    call addComment(opt_arr, shortname)

! Define dummy offsets if no OFFSETS extension in events file

    if (.not. onoffvar) then
       off2_tab = addTable(e4_set, offset_ext, 1, "Dummy offset/variance")
       call setAttributes(attributable(off2_tab), attributable(ev_tab))
       call deleteAttribute(off2_tab,"EXTVER")
       call setAttribute(off2_tab, "DATATYPE", "OFFVAR.EL")
       call setAttribute(off2_tab, "CCDVAR", 1., comment="Dummy variance")
       offx_col = addColumn(off2_tab, "COLOFFST", Integer16,  &
                  comment="Dummy row offsets", dimensions=(/x_size_window/))
       offsetx => int16VectorData(offx_col)
       offy_col = addColumn(off2_tab, "ROWOFFST", Integer16,  &
                  comment="Dummy column offsets", dimensions=(/y_size_window/))
       offsety => int16VectorData(offy_col)
       offsetx = 50
       offsety = 50
       shortname = "Dummy offsets are added to the background map"
    else
       shortname = "Offsets are added back to the background map"
    endif
    call addComment(opt_arr, shortname)

    e4img => real32Array2Data(opt_arr)
    e4img = optbkg(1:EMOS_MAX_X,1:EMOS_MAX_Y)
    do irow = 0, EMOS_MAX_X-1
       e4img(irow,:) = e4img(irow,:) + offsetx(underx+irow,0)
    enddo
    do irow = 0, EMOS_MAX_Y-1
       e4img(:,irow) = e4img(:,irow) + offsety(undery+irow,0)
    enddo
    call release(opt_arr)
    if (.not. onoffvar)  call release(off2_tab)
    wght_arr = addArray(e4_set, "CCD_WGHT", Integer32,  &
        (/EMOS_MAX_X,EMOS_MAX_Y/), "weights for background map")
    shortname = "Number of pixels contributing to each value in background map"
    call addComment(wght_arr, shortname)
    e4wght => int32Array2Data(wght_arr)
    e4wght = sigmaper(1:EMOS_MAX_X,1:EMOS_MAX_Y)
    call release(wght_arr)

! Initialize the time-series

    if (timebin > epsilon) then
       time0 = int(minval(timetag)/timebin) * timebin
       nbin = int((maxval(timetag)-time0)/timebin) + 1
    else
       write(errstr,*) "the time bin ",timebin," for ",trim(backgrnd), &
          " is too small. Should be larger than ", epsilon,  &
          ". Background time series will be empty."
       call warning("ccdbkg15")
       nbin = 0
    endif
    if (nbin > nbinmax) then
       write(errstr,*) "the number of time bins ", nbin, " for ",  &
          trim(backgrnd), "is too large. Should be smaller than ",  &
          nbinmax, ". Background time series will be empty."
       call warning("ccdbkg13")
       nbin = 0
    endif

    e4_tab = addTable(e4_set, "CCD_VAR", nbin, "time variations of background")
    call setAttributes(attributable(e4_tab), attributable(ev_tab))
    call deleteAttribute(e4_tab,"DATATYPE")
    call deleteAttribute(e4_tab,"EXTVER")
    shortname = "Binning for the time series"
    call setAttribute(e4_tab,"TIMEDEL",timebin,"s",comment=shortname)
    shortname = "Start time for the time series"
    call setAttribute(e4_tab,"TIMEZERO",time0,"s",shortname)
    shortname = "Offsets are subtracted in CCD_VAR"
    call addComment(e4_tab, shortname)

! Calculate the time-series

    if (nbin > 0) then
       te4_col = addColumn(e4_tab, "BKGCHARGE", Real32, "CHANNEL",  &
                           comment="Averaged background in time bin")
       timee4 => real32Data(te4_col)
       tper_col = addColumn(e4_tab, "BKGWEIGHT", Integer32, "",  &
                           comment="Number of events in time bin")
       timeper => int32Data(tper_col)
       timee4 = 0.
       timeper = 0
       do irow = 0, numberOfRows(ev_tab)-1

! Same test as in ccdBkg

          if (ener1(irow) < minadubkg .or. iand(quality(irow),testflag) > 0  &
              .or. pixelInE2(pattern(irow)) > 1  &
              .or. peripix(irow) > perithresh)  cycle
          ibin = int((timetag(irow)-time0)/timebin)
           timee4(ibin) =  timee4(ibin) + ener4(irow)
          timeper(ibin) = timeper(ibin) + 16-peripix(irow)
       enddo
       where (timeper /= 0)
          timee4 = timee4 / timeper
       endwhere
       call release(te4_col)
       call release(tper_col)
    endif
    call release(e4_tab)

! Copy the OFFSETS extension for information

    if (onoffvar)  call copyBlock(e4_set, block(off_tab), offset_ext)
    call release(e4_set)

  end subroutine writeBkg

! Read the required columns into global variables

  subroutine readColumns(ev_tab, pha_col, pi_col)

! ev_tab    : Handle to the input/output event table
! pha_col   : Handle to the PHA column
!  pi_col   : Handle to the  PI column

    type(TableT), intent(in)   :: ev_tab
    type(ColumnT), intent(out) :: pha_col, pi_col

    type(ColumnT) :: patt_col, flag_col, x_col, y_col, peri_col,  &
                     e1_col, e4_col, time_col

    integer :: phamax, phamin

    if (hasColumn(ev_tab, EVATT_COLUMN_NAME)) then
       flag_col = column(ev_tab, EVATT_COLUMN_NAME, Modify)
       quality => int32Data(flag_col)
    else
       flag_col = addColumn(ev_tab, EVATT_COLUMN_NAME, Integer32, "",  &
                            comment="quality flag")
       quality => int32Data(flag_col)
       quality = 0
       write(errstr,*) "Added column ", EVATT_COLUMN_NAME, " set to 0."
       call message(AppMsg,Sparse)
    endif

    x_col = column(ev_tab, rawx, Read)
    posx => int16Data(x_col)
    if (imageMode .or. rimagMode) then
       y_col = column(ev_tab, rawy, Read)
       posy => int16Data(y_col)
    else if (timeMode .or. ctimeMode) then
       if (oncticorr) then     ! CAL_mosCTIcorrection formally demands posy
          y_col = column(ev_tab, rawy, Read)
          posy => int16Data(y_col)
       endif

! Flag events of unknown abscissa (COMPRESSED TIMING) with x=-1

       if (ctimeMode) then
          where (posx == -1)
             quality = ior(quality,EVATT_UNKNOWN_RAWX_PATTERN)
          endwhere
          call setEventAttribute(ev_tab,EVATT_UNKNOWN_RAWX_PATTERN)
       endif
    endif

    patt_col = column(ev_tab, patt, Read)
    pattern => int8Data(patt_col)
    peri_col = column(ev_tab, peri, Modify)
    peripix => int8Data(peri_col)

    if (imageMode) then
        e4_col = column(ev_tab, ene4, Read)
       ener4 => int16Data(e4_col)
    endif

    if (onmakee .or. oncticorr .or. onenergy) then
       if (hasColumn(ev_tab, trim(pha))) then
          pha_col = column(ev_tab, pha, Modify)
       else
          pha_col = addColumn(ev_tab, pha, Real32, "CHANNELS",  &
                       comment="total charge of the event (uncorrected)")
          write(errstr,*) "Added column ", trim(pha)
          call message(AppMsg,Sparse)
       endif
       phaval => real32Data(pha_col)
    endif

    if (imageMode .or. rimagMode) then
        e1_col = column(ev_tab, ene1, Read)
       ener1 => int16Data(e1_col)
       if (onmakee) then

! Read TDMIN/TDMAX from ENERGYE1

          if (hasAttribute(e1_col,"TDMIN")) then
             phamin = int32Attribute(e1_col,"TDMIN")
          else
             phamin = 0
          endif
          if (hasAttribute(e1_col,"TDMAX")) then
             phamax = int32Attribute(e1_col,"TDMAX")
          else
             phamax = 4095
          endif
          call setAttribute(pha_col, "TDMIN", phamin,  &
                            comment="Actual minimum for "//pha)
          call setAttribute(pha_col, "TDMAX", phamax,  &
                            comment="Actual maximum for "//pha)
       endif
    endif

    if (oncticorr .or. onenergy) then
       if (hasColumn(ev_tab, trim(pi))) then
          pi_col = column(ev_tab, pi, Modify)
       else
          pi_col = addColumn(ev_tab, pi, Real32, "CHANNELS",  &
                     comment="total charge of the event (gain corrected)")
          write(errstr,*) "Added column ", trim(pi)
          call message(AppMsg,Sparse)
       endif
       pival => real32Data(pi_col)
    endif

    if (oncticorr .or. onenergy .or. one4bkg) then
       time_col = column(ev_tab, time, Read)
       timetag => real64Data(time_col)
    endif

  end subroutine readColumns

! Read the pattern library, set the 2s and the central to 0
! Deduce the number of pixels making E2

  subroutine getpixelInE2(pixelInE2, patabove2)

    integer(kind=int8), dimension(0:31), intent(out)        :: pixelInE2
    integer(kind=int8), dimension(0:31,-1:1,-1:1),  &
                        intent(out), optional               :: patabove2

    integer(int8), dimension(:,:,:), pointer      :: patterns
    integer(int8), dimension(0:31,-1:1,-1:1)      :: patabove
    integer(int16)                                :: eduThreshold

    call CAL_getEventPatterns(patterns, eduThreshold)
    patabove = patterns(:,1:3,1:3)
! Can't release it
!   call CAL_releaseMemory(patterns)
    patabove(:,0,0) = 0
    where (patabove > 1)
       patabove = 0
    endwhere
    pixelInE2 = sum(sum(patabove,3),2)

    if (present(patabove2))  patabove2 = patabove

  end subroutine getpixelInE2

! Initialize the CAL and the image size, set randomisation

  subroutine initCAL(ev_tab, fovMap)

! ev1_tab : handle to the input table
! fovMap  : Output of CAL_getFOVmap

    type(TableT), intent(in) :: ev_tab
    integer(kind=int16), dimension(:,:), pointer :: fovMap

    type(ModeParameterDataType)  :: modeParameters

    call CAL_setState(block(ev_tab))

    call CAL_setState(randomize=e1random)

! Get under/overscans

    call CAL_getModeParameters(modeParameters)
    underx = modeParameters%underscanX
    overx  = modeParameters%overscanX
    undery = modeParameters%underscanY
    overy  = modeParameters%overscanY
    x_size_window = EMOS_MAX_X + underx + overx
    y_size_window = EMOS_MAX_Y + undery + overy

    windX0 = int32Attribute(ev_tab,"WINDOWX0")
    windDX = int32Attribute(ev_tab,"WINDOWDX")
    windY0 = int32Attribute(ev_tab,"WINDOWY0")
    windDY = int32Attribute(ev_tab,"WINDOWDY")
    write(errstr,*) " WINDOWX0 =",windX0,"     WINDOWY0 =",windY0
    call message(AppMsg,Verbose)

! Flag events outside FOV

    if ((imageMode .or. rimagMode) .and. (onevflag .or. onfillbkg))  &
       call CAL_getFOVmap(fovMap, int(windX0,int16), int(windY0,int16),  &
                                  int(windDX,int16), int(windDY,int16))

  end subroutine initCAL

! Parameter reading

  subroutine getParamvalues(ev_tab, off_tab, diagnos, backgrnd, timebin,  &
                            minadubkg, perithresh, maxevtfill,  &
                            e3cen, e4cen, e3rad, e4rad, e3ang)

! ev_tab     : Handle to the input/output event table
! off_tab    : Handle to the offset table
! diagnos    : Name of the diagnostic map file for CCDBKG
! backgrnd   : Name of the E4 map and timeseries file
! timebin    : Time bin for E4 timeseries
! minadubkg  : Min E1 for building E4 map
! perithresh : Max PERIPIX for building E4 map
! maxevtfill : Max number of evts for averaging E4 map
! e3cen      : Center of ellipse along E3/E1
! e4cen      : Center of ellipse along E4/E1
! e3rad      : Radius of ellipse along first axis
! e4rad      : Radius of ellipse along second axis
! e3ang      : Angle of first axis with E3 axis

    character(len=*), intent(out) :: diagnos, backgrnd
    real,             intent(out) :: timebin, e3cen, e4cen, e3rad, e4rad, e3ang
    integer,          intent(out) :: minadubkg, perithresh, maxevtfill
    type(TableT),     intent(out) :: ev_tab, off_tab

    type(DataSetT) :: ev_set
    type(ColumnT)  :: offx_col, offy_col
    character(len=1024) :: event_in, event_out
    integer        :: nellipse

    event_in = stringParameter(inevent)
    if (len_trim(event_in) == 0) &
       call error("getParamvalues01","Blank input event file name")
    write(errstr,*) "Input file : ",trim(event_in)
    call message(AppMsg,Sparse)

    if (booleanParameter(newoutput)) then
       event_out = stringParameter(outevent)
       call copyDataSet(event_in, event_out)
       write(errstr,*) "Input file ", trim(event_in), " copied to ", &
                       trim(event_out)
       call message(AppMsg,Sparse)
    else
       event_out = event_in
       call message(AppMsg,Sparse,"The input file will be overwritten")
    endif
    ev_set = dataSet(event_out, Modify)
    ev_tab = table(ev_set, event_ext)

    call getMode(ev_tab, imageMode, timeMode, rimagMode, ctimeMode)

    onexpos = hasBlock(ev_set, expos_ext)
    if (.not. onexpos) then
       write(errstr,*) "no ", expos_ext, " extension in input eventfile.",  &
                       " The keywords won't be propagated after evlistcomb."
       call warning("getParamvalues10")
    endif

    if (numberOfRows(ev_tab) > 0) then
       write(errstr,*) numberOfRows(ev_tab), " events in file"
       call message(AppMsg,Sparse)
    else
       write(errstr,*) "the event file is empty."
       call warning("getParamvalues20")
       return
    endif

    if (.not. wasDone(block(ev_tab),changexy))  then
       write(errstr,*) "PUT_XY must have been applied to the data first.",  &
                       " Launch emevents with ",changexy,"=yes"
       call error("getParamvalues03")
    endif

! Read offset/variance extension

    onoffvar = hasBlock(ev_set, offset_ext)
    if (onoffvar) then
       off_tab = table(ev_set, offset_ext)
       offx_col = column(off_tab, "COLOFFST", Read)
       offsetx => int16VectorData(offx_col)
       offy_col = column(off_tab, "ROWOFFST", Read)
       offsety => int16VectorData(offy_col)
    endif

    onevflag  = booleanParameter(evflag_on)
    if (onevflag .and. wasDone(block(ev_tab),evflag_on)) then
       onevflag = .false.
       call warning("getParamvalues11",  &
                    "EVFLAG cannot be applied twice. Not repeated")
    endif
    if (onevflag) then
       call message(AppMsg,Sparse,"EVFLAG will be applied")
       if (onoffvar) then
          write(errstr,*) "Offsets will be taken from the ", offset_ext,  &
                          " extension."
       else
          write(errstr,*) "No offset/variance extension was found. ", &
                          "All lines/columns assumed OK"
       endif
       call message(AppMsg,Sparse)
    endif

    oncticorr = booleanParameter(cticorr_on)
    if (oncticorr)  call message(AppMsg,Sparse,"CTI_CORR will be applied")

    ondark   = booleanParameter(withdark)
    ondiag   = booleanParameter(withdiag)
    if (ondark .and. ondiag) then
       ondark = .false.
       write(errstr,*) "the present version of CCDBKG ", &
             "does not support a CCD dark frame and a diagnostic map ", &
             "at the same time. Dark frame not applied"
       call warning("getParamvalues15")
    endif

! Dark frame correction makes sense only in IMAGING mode

    if ((ondark .or. ondiag) .and. .not. (imageMode .or. rimagMode)) then
       ondark = .false.
       ondiag = .false.
       call warning("getParamvalues12",  &
        "Dark frame correction may be applied only in IMAGING mode. Not done.")
    endif
    if (ondiag) then
       diagnos = stringParameter(diagfile)
       write(errstr,*) "Dark frame correction will be applied ",  &
                       "with the diagnostic map ", diagnos
    else if (ondark) then
       write(errstr,*) "Dark frame correction will use the CCF"
    endif
    call message(AppMsg,Sparse)

    one4bkg  = booleanParameter(ccdbkg_on)
    outbkg   = .false.

    if (one4bkg) then
       if (hasAttribute(ev_set,'OBS_MODE')) then
          if (stringAttribute(ev_set,'OBS_MODE') == 'SLEW') then
             one4bkg = .false.
             write(errstr,*) "Computing the CCD background by averaging",  &
                             " E4 is not possible for SLEW data. Not done."
             call warning("getParamvalues26")
          endif
       endif
       if (.not. imageMode) then
          one4bkg = .false.
          write(errstr,*) "Computing the CCD background by averaging E4",  &
                          " is not possible in REDUCED IMAGING mode. Not done."
          call warning("getParamvalues13")
       endif
    endif

    onfillbkg = .false.
    if (one4bkg) then
       call message(AppMsg,Sparse,  &
                    "Energies will be corrected using E4 information")
       minadubkg = intParameter(minenergy)
       write(errstr,*) "Events with ENERGYE1 >=", minadubkg,  &
                       " will be used to build the background map."
       call message(AppMsg,Sparse)
       perithresh = intParameter(maxperipix)
       write(errstr,*) "Events with PERIPIX <=", perithresh,  &
                       " will be used to build the background map."
       call message(AppMsg,Sparse)
       onfillbkg = booleanParameter(fillbkg_on)
       if (onfillbkg) then
          maxevtfill = intParameter(fillthresh)
          write(errstr,*) "Background data will be filled by ",  &
                          "row and column averaging when less than ",  &
                          maxevtfill, " events around."
          call message(AppMsg,Sparse)
       endif
       outbkg = booleanParameter(savebkg)
       if (outbkg) then
          backgrnd = stringParameter(bkgfile)
          timebin = realParameter(tbin)
          write(errstr,*) "Background data will be written to file ", &
                          trim(backgrnd)
          call message(AppMsg,Sparse)
          write(errstr,*) "Time bin for time series :", timebin
          call message(AppMsg,Sparse)
       endif
    endif

    e3e4reject = booleanParameter(bade3e4_on)
    if ( e3e4reject ) then
       if (.not. imageMode) then
          e3e4reject = .false.
          write(errstr,*) "The ",bade3e4_on,"=yes option may be applied ",  &
                          "only in IMAGING mode. Not done."
          call warning("getParamvalues22")
       else
          nellipse = parameterCount(e3e4ellipse)
          if (nellipse < 4) then
             write(errstr,*) "The number of ellipse parameters in ",  &
                             e3e4ellipse, nellipse, " should be at least 4."
             call error("getParamvalues02")
          endif
          e3cen = realParameter(e3e4ellipse,0)
          e4cen = realParameter(e3e4ellipse,1)
          e3rad = realParameter(e3e4ellipse,2)
          e4rad = realParameter(e3e4ellipse,3)
          e3ang = 0.
          if (nellipse > 4)  e3ang = realParameter(e3e4ellipse,4)
          write(errstr,*) "Events with (E3/E1,E4/E1) outside of an ellipse ", &
               "centered on", e3cen, e4cen, " and of radii", e3rad, e4rad,  &
              " at angle", e3ang, " radians with the E3 axis will be rejected."
          call message(AppMsg,Verbose)
          if (nellipse > 5) then
             write(errstr,*) "The number of ellipse parameters in ",  &
                             e3e4ellipse, nellipse, " should be at most 5. ", &
                             "Additional values are ignored."
             call warning("getParamvalues16")
          endif
       endif
    endif

    onmakee = booleanParameter(makee_on)
    if (onmakee) then
       e1random = booleanParameter(randome1)
       if (e1random)  call message(AppMsg,Sparse,"PHA/PI will be randomized.")
       if ((imageMode.or.rimagMode) .and. (.not. one4bkg))  &
          call warning("getParamvalues17",  &
                   "CCDBKG not activated. MAKE_E will assume zero background.")
    endif

    onenergy  = booleanParameter(energy_on)
    if (onenergy .and.  &
        .not.(wasDone(block(ev_tab),makee_on).or.onmakee)) then
       onenergy = .false.
       call warning("getParamvalues18", &
          "ENERGY cannot be applied before MAKE_E. Not applied.")
    endif
    if (onenergy)  call message(AppMsg,Sparse,"ENERGY will be applied.")

  end subroutine getParamvalues

end module emenergy_mod
